@@ -3,8 +3,8 @@
 /*
  * Squashfs
  *
- * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011
- * 2012, 2013, 2014, 2019, 2021, 2022, 2023
+ * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011,
+ * 2012, 2013, 2014, 2019, 2021, 2022, 2023, 2024, 2025
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -24,6 +24,8 @@
  * mksquashfs.h
  *
  */
+#include <pthread.h>
+#include <unistd.h>
 
 struct dir_info {
 	char			*pathname;
@@ -46,26 +48,31 @@ struct dir_ent {
 	struct dir_info		*dir;
 	struct dir_info		*our_dir;
 	struct dir_ent		*next;
+	struct dir_ent		*reader_next;
 };
 
 struct inode_info {
 	struct stat		buf;
 	struct inode_info	*next;
 	struct pseudo_dev	*pseudo;
-	struct tar_file		*tar_file;
+	union {
+		struct tar_file		*tar_file;
+		struct file_info	*file;
+	};
 	struct pseudo_xattr	*xattr;
 	squashfs_inode		inode;
 	unsigned int		inode_number;
 	unsigned int		nlink;
 	char			dummy_root_dir;
 	char			type;
-	char			read;
 	char			root_entry;
 	char			no_fragments;
 	char			always_use_fragments;
 	char			noD;
 	char			noF;
 	char			tarfile;
+	char			read;
+	char			scanned;
 	char			symlink[0];
 };
 
@@ -132,31 +139,22 @@ struct append_file {
 #define SQUASHFS_FWRITEQ_MEM 4
 
 /*
- * Lowest amount of physical memory considered viable for Mksquashfs
- * to run in Mbytes
+ * Default memory for when Mksquashfs cannot get the amount of
+ * memory in the system
  */
-#define SQUASHFS_LOWMEM 64
+#define SQUASHFS_UNKNOWN_MEM 1024
 
 /* offset of data in compressed metadata blocks (allowing room for
  * compressed size */
 #define BLOCK_OFFSET 2
 
-#ifdef REPRODUCIBLE_DEFAULT
-#define NOREP_STR
-#define REP_STR " (default)"
-#define REP_DEF 1
-#else
-#define NOREP_STR " (default)"
-#define REP_STR
-#define REP_DEF 0
-#endif
-
 /* in memory directory data */
 #define I_COUNT_SIZE		128
 #define DIR_ENTRIES		32
 #define INODE_HASH_SIZE		65536
-#define INODE_HASH_MASK		(INODE_HASH_SIZE - 1)
-#define INODE_HASH(dev, ino)	(ino & INODE_HASH_MASK)
+#define INO_MASK(ino, shift)	((((unsigned long long) ino) >> shift) & 0xffff)
+#define INODE_HASH(dev, ino)	((INO_MASK(ino, 0) + INO_MASK(ino, 16) + \
+				INO_MASK(ino, 32) + INO_MASK(ino, 48)) & 0xffff)
 
 struct cached_dir_index {
 	struct squashfs_dir_index	index;
@@ -167,14 +165,15 @@ struct directory {
 	unsigned int		start_block;
 	unsigned int		size;
 	unsigned char		*buff;
-	unsigned char		*p;
+	unsigned int		offset;
 	unsigned int		entry_count;
-	unsigned char		*entry_count_p;
+	unsigned int		entry_count_offset;
 	unsigned int		i_count;
 	unsigned int		i_size;
 	struct cached_dir_index	*index;
-	unsigned char		*index_count_p;
+	unsigned int		index_count_offset;
 	unsigned int		inode_number;
+	char			have_dir_header;
 };
 
 /* exclude file handling */
@@ -210,11 +209,31 @@ struct old_root_entry_info {
  * See READ(2) */
 #define MAXIMUM_READ_SIZE 0x7ffff000
 
-extern int sleep_time;
-extern struct cache *reader_buffer, *fragment_buffer, *reserve_cache;
-extern struct cache *bwriter_buffer, *fwriter_buffer;
-extern struct queue *to_reader, *to_deflate, *to_writer, *from_writer,
-	*to_frag, *locked_fragment, *to_process_frag;
+static inline int get_pathmax()
+{
+	int path_max;
+
+#ifdef PATH_MAX
+	path_max = PATH_MAX;
+#else
+	path_max = pathconf(".", _PC_PATH_MAX);
+	if(path_max <= 0)
+		path_max = 4096;
+#endif
+	/* limit to no more than 64K */
+	if(path_max > 65536)
+		path_max = 65536;
+
+	return path_max;
+}
+
+extern struct cache *fragment_buffer, *reserve_cache;
+extern struct cache *fwriter_buffer;
+extern struct queue_cache *bwriter_buffer;
+extern struct queue *to_reader, *to_writer, *from_writer, *to_frag,
+       *locked_fragment;
+extern struct queue_cache *to_deflate;
+extern struct read_queue *to_process_frag;
 extern struct append_file **file_mapping;
 extern struct seq_queue *to_main, *to_order;
 extern pthread_mutex_t fragment_mutex, dup_mutex;
@@ -235,7 +254,7 @@ extern struct dir_info *root_dir;
 extern struct pathnames *paths;
 extern int tarfile;
 extern int root_mode_opt;
-extern mode_t root_mode;
+extern struct mode_data *root_mode;
 extern int root_time_opt;
 extern unsigned int root_time;
 extern int root_uid_opt;
@@ -244,13 +263,13 @@ extern int root_gid_opt;
 extern unsigned int root_gid;
 extern struct inode_info *inode_info[INODE_HASH_SIZE];
 extern int quiet;
-extern int sequence_count;
 extern int pseudo_override;
 extern int global_uid_opt;
 extern unsigned int global_uid;
 extern int global_gid_opt;
 extern unsigned int global_gid;
-extern int sleep_time;
+extern int global_dir_mode_opt;
+extern struct mode_data *global_dir_mode;
 
 extern int read_fs_bytes(int, long long, long long, void *);
 extern void add_file(long long, long long, long long, unsigned int *, int,
@@ -261,7 +280,7 @@ extern unsigned int get_guid(unsigned int);
 extern long long read_bytes(int, void *, long long);
 extern unsigned short get_checksum_mem(char *, int);
 extern int reproducible;
-extern void *reader(void *arg);
+extern void *initial_reader(void *arg);
 extern squashfs_inode create_inode(struct dir_info *dir_info,
 	struct dir_ent *dir_ent, int type, long long byte_size,
 	long long start_block, unsigned int offset, unsigned int *block_list,
@@ -281,5 +300,6 @@ extern char *subpathname(struct dir_ent *dir_ent);
 extern struct dir_info *scan1_opendir(char *pathname, char *subpath, unsigned int depth);
 extern squashfs_inode do_directory_scans(struct dir_ent *dir_ent, int progress);
 extern struct inode_info *lookup_inode(struct stat *buf);
+extern struct inode_info *lookup_inode_flag(struct stat *buf, int have_time);
 extern int exec_date(char *, unsigned int *);
 #endif
